"""Input/output functions."""
try:
    import pandas as pd
except ImportError as e:
    raise ImportError(
        "pandas is required to use the i/o utils (mlcolvar.utils.io)\n", e
    )

import numpy as np
import torch
import os
import urllib.request
from typing import Union

from mlcolvar.data import DictDataset

__all__ = ["load_dataframe", "plumed_to_pandas", "create_dataset_from_files"]


def is_plumed_file(filename):
    """
    Check if given file is in PLUMED format.

    Parameters
    ----------
    filename : string, optional
        PLUMED output file

    Returns
    -------
    bool
        wheter is a plumed output file
    """
    headers = pd.read_csv(filename, sep=" ", skipinitialspace=True, nrows=0)
    is_plumed = True if " ".join(headers.columns[:2]) == "#! FIELDS" else False
    return is_plumed


def plumed_to_pandas(filename="./COLVAR"):
    """
    Load a PLUMED file and save it to a dataframe.

    Parameters
    ----------
    filename : string, optional
        PLUMED output file

    Returns
    -------
    df : DataFrame
        Collective variables dataframe
    """
    skip_rows = 1
    # Read header
    headers = pd.read_csv(filename, sep=" ", skipinitialspace=True, nrows=0)
    # Discard #! FIELDS
    headers = headers.columns[2:]
    # Load dataframe and use headers for columns names
    df = pd.read_csv(
        filename,
        sep=" ",
        skipinitialspace=True,
        header=None,
        skiprows=range(skip_rows),
        names=headers,
        comment="#",
    )

    return df


def load_dataframe(
    file_names, start=0, stop=None, stride=1, delete_download=True, **kwargs
):
    """Load dataframe(s) from file(s). It can be used also to open files from internet (if the string contains http).
    In case of PLUMED colvar files automatically handles the column names, otherwise it is just a wrapper for pd.load_csv function.

    Parameters
    ----------
    filenames : str or list[str]
        filenames to be loaded
    start: int, optional
        read from this row, default 0
    stop: int, optional
        read until this row, default None
    stride: int, optional
        read every this number, default 1
    delete_download: bool, optinal
        whether to delete the downloaded file after it has been loaded, default True.
    kwargs:
        keyword arguments passed to pd.load_csv function

    Returns
    -------
    pandas.DataFrame
        Dataframe

    Raises
    ------
    TypeError
        if data is not a valid type
    """

    # if it is a single string
    if type(file_names) == str:
        file_names = [file_names]
    elif type(file_names) != list:
        raise TypeError(
            f"only strings or list of strings are supported, not {type(file_names)}."
        )

    # list of file_names
    df_list = []
    for i, filename in enumerate(file_names):
        # check if filename is an url
        download = False
        if "http" in filename:
            download = True
            url = filename
            filename = "tmp_" + filename.split("/")[-1]
            urllib.request.urlretrieve(url, filename)

        # check if file is in PLUMED format
        if is_plumed_file(filename):
            df_tmp = plumed_to_pandas(filename)
            df_tmp["walker"] = [i for _ in range(len(df_tmp))]
            df_tmp = df_tmp.iloc[start:stop:stride, :]
            df_list.append(df_tmp)

        # else use read_csv with optional kwargs
        else:
            df_tmp = pd.read_csv(filename, **kwargs)
            df_tmp["walker"] = [i for _ in range(len(df_tmp))]
            df_tmp = df_tmp.iloc[start:stop:stride, :]
            df_list.append(df_tmp)

        # delete temporary data if necessary
        if download:
            if delete_download:
                os.remove(filename)
            else:
                print(f"downloaded file ({url}) saved as ({filename}).")

        # concatenate
        df = pd.concat(df_list)
        df.reset_index(drop=True, inplace=True)

    return df


def create_dataset_from_files(
    file_names: Union[list, str],
    folder: str = None,
    create_labels: bool = None,
    load_args: list = None,
    filter_args: dict = None,
    modifier_function=None,
    return_dataframe: bool = False,
    verbose: bool = True,
    **kwargs,
):
    """
    Initialize a dataset from (a list of) files. Suitable for supervised/unsupervised tasks.

    Parameters
    ----------
    file_names : list
        Names of files from which import the data
    folder : str, optional
        Common path for the files to be imported, by default None. If set, filenames become 'folder/file_name'.
    create_labels: bool, optional
        Assign a label to each file, default True if more than a file is given, otherwise False
    load_args: list[dict], optional
        List of dictionaries with the arguments passed to load_dataframe function for each file (keys: start,stop,stride and pandas.read_csv options), by default None
    filter_args: dict, optional
        Dictionary of arguments which are passed to df.filter() to select descriptors (keys: items, like, regex), by default None
        Note that 'time' and '*.bias' columns are always discarded.
    return_dataframe : bool, optional
        Return also the imported Pandas dataframe for convenience, by default False
    modifier_function : function, optional
        Function to be applied to the input data, by default None.
    verbose : bool, optional
        Print info on the datasets, by default True
    kwargs : optional
        args passed to mlcolvar.utils.io.load_dataframe

    Returns
    -------
    torch.Dataset
        Torch labeled dataset of the given data
    optional, pandas.Dataframe
        Pandas dataframe of the given data #TODO improve

    See also
    --------
    mlcolvar.utils.io.load_dataframe
        Function that is used to load the files

    """
    if isinstance(file_names, str):
        file_names = [file_names]

    num_files = len(file_names)

    # set file paths
    if folder is not None:
        file_names = [os.path.join(folder, fname) for fname in file_names]

    # check if per file args are given, otherwise set to {}
    if load_args is None:
        load_args = [{} for _ in file_names]
    else:
        if (not isinstance(load_args, list)) or (len(file_names) != len(load_args)):
            raise TypeError(
                "load_args should be a list of dictionaries of arguments of same length as file_names. If you want to use the same args for all file pass them directly as **kwargs."
            )

    # check if create_labels if given, otherwise set it to True if more than one file is given
    if create_labels is None:
        create_labels = False if len(file_names) == 1 else True

    # initialize pandas dataframe
    df = pd.DataFrame()

    # load data
    for i in range(num_files):
        df_tmp = load_dataframe(file_names[i], **load_args[i], **kwargs)

        # add label in the dataframe
        if create_labels:
            df_tmp["labels"] = i
        if verbose:
            print(f"Class {i} dataframe shape: ", np.shape(df_tmp))

        # update collective dataframe
        df = pd.concat([df, df_tmp], ignore_index=True)

    # filter inputs
    df_data = df.filter(**filter_args) if filter_args is not None else df.copy()
    df_data = df_data.filter(regex="^(?!.*labels)^(?!.*time)^(?!.*bias)^(?!.*walker)")

    if verbose:
        print(f"\n - Loaded dataframe {df.shape}:", list(df.columns))
        print(f" - Descriptors {df_data.shape}:", list(df_data.columns))

    # apply transformation
    if modifier_function is not None:
        df_data = df_data.apply(modifier_function)

    # create DictDataset
    dictionary = {"data": torch.Tensor(df_data.values)}
    if create_labels:
        dictionary["labels"] = torch.Tensor(df["labels"].values)
    dataset = DictDataset(dictionary)

    if return_dataframe:
        return dataset, df
    else:
        return dataset


def test_datasetFromFile():
    # Test with unlabeled dataset
    torch_dataset, pd_dataframe = create_dataset_from_files(
        file_names=["state_A.dat", "state_B.dat", "state_C.dat"],
        folder="mlcolvar/tests/data",
        create_labels=False,
        load_args=None,
        filter_args=None,
        return_dataframe=True,
        start=0,  # kwargs to load_dataframe
        stop=5,
        stride=1,
    )

    # Test no regex on two states
    create_dataset_from_files(
        file_names=["state_A.dat", "state_B.dat"],
        folder="mlcolvar/tests/data",
        create_labels=True,
        load_args=None,
        filter_args=None,
        return_dataframe=True,
        start=0,  # kwargs to load_dataframe
        stop=5,
        stride=1,
    )

    # Test with filter regex on two states
    dataset = create_dataset_from_files(
        file_names=["state_A.dat", "state_B.dat"],
        folder="mlcolvar/tests/data",
        create_labels=True,
        load_args=None,
        filter_args={"regex": "n|o"},
        return_dataframe=False,
        start=0,  # kwargs to load_dataframe
        stop=5,
        stride=1,
    )

    def test_modifier(x):
        return x**2

    # Test with filter regex on two states with modifier
    create_dataset_from_files(
        file_names=["state_A.dat", "state_B.dat"],
        folder="mlcolvar/tests/data",
        create_labels=True,
        load_args=None,
        filter_args={"regex": "n|o"},
        modifier_function=test_modifier,
        return_dataframe=True,
        start=0,  # kwargs to load_dataframe
        stop=5,
        stride=1,
    )


if __name__ == "__main__":
    test_datasetFromFile()



import torch
import lightning
from mlcolvar.cvs import BaseCV
from mlcolvar.core import FeedForward, Normalization
from mlcolvar.core.stats import TICA
from mlcolvar.core.loss import ReduceEigenvaluesLoss

__all__ = ["DeepTICA"]


class DeepTICA(BaseCV, lightning.LightningModule):
    """Neural network-based time-lagged independent component analysis (Deep-TICA).

    It is a non-linear generalization of TICA in which a feature map is learned by a
    neural network optimized as to maximize the eigenvalues of the transfer operator,
    approximated by TICA. The method is described in [1]_. Note that from the point of view
    of the architecture DeepTICA is similar to the SRV [2] method.

    **Data**: for training it requires a DictDataset with the keys 'data' (input at time t)
    and 'data_lag' (input at time t+lag), as well as the corresponding 'weights' and
    'weights_lag' which will be used to weight the time correlation functions.
    This can be created with the helper function `create_timelagged_dataset`.

    **Loss**: maximize TICA eigenvalues (ReduceEigenvaluesLoss)

    References
    ----------
    .. [1] L. Bonati, G. Piccini, and M. Parrinello, “ Deep learning the slow modes for
        rare events sampling,” PNAS USA 118, e2113533118 (2021)
    .. [2] W. Chen, H. Sidky, and A. L. Ferguson, “ Nonlinear discovery of slow molecular
        modes using state-free reversible vampnets,” JCP 150, 214114 (2019).

    See also
    --------
    mlcolvar.core.stats.TICA
        Time Lagged Indipendent Component Analysis
    mlcolvar.core.loss.ReduceEigenvalueLoss
        Eigenvalue reduction to a scalar quantity
    mlcolvar.utils.timelagged.create_timelagged_dataset
        Create dataset of time-lagged data.
    """

    BLOCKS = ["norm_in", "nn", "tica"]

    def __init__(self, layers: list, n_cvs: int = None, options: dict = None, **kwargs):
        """
        Define a Deep-TICA CV, composed of a neural network module and a TICA object.
        By default a module standardizing the inputs is also used.

        Parameters
        ----------
        layers : list
            Number of neurons per layer
        n_cvs : int, optional
            Number of cvs to optimize, default None (= last layer)
        options : dict[str, Any], optional
            Options for the building blocks of the model, by default {}.
            Available blocks: ['norm_in','nn','tica'].
            Set 'block_name' = None or False to turn off that block
        """
        super().__init__(
            in_features=layers[0],
            out_features=n_cvs if n_cvs is not None else layers[-1],
            **kwargs,
        )

        # =======   LOSS  =======
        # Maximize the squared sum of all the TICA eigenvalues.
        self.loss_fn = ReduceEigenvaluesLoss(mode="sum2")

        # ======= OPTIONS =======
        # parse and sanitize
        options = self.parse_options(options)

        # ======= BLOCKS =======

        # initialize norm_in
        o = "norm_in"
        if (options[o] is not False) and (options[o] is not None):
            self.norm_in = Normalization(self.in_features, **options[o])

        # initialize nn
        o = "nn"
        self.nn = FeedForward(layers, **options[o])

        # initialize lda
        o = "tica"
        self.tica = TICA(layers[-1], n_cvs, **options[o])

    def forward_nn(self, x: torch.Tensor) -> torch.Tensor:
        if self.norm_in is not None:
            x = self.norm_in(x)
        x = self.nn(x)
        return x

    def set_regularization(self, c0_reg=1e-6):
        """
        Add identity matrix multiplied by `c0_reg` to correlation matrix C(0) to avoid instabilities in performin Cholesky and .

        Parameters
        ----------
        c0_reg : float
            Regularization value for C_0.
        """
        self.tica.reg_C_0 = c0_reg

    def training_step(self, train_batch, batch_idx):
        """Compute and return the training loss and record metrics.
        1) Calculate the NN output
        2) Remove average (inside forward_nn)
        3) Compute TICA
        """
        # =================get data===================
        x_t = train_batch["data"]
        x_lag = train_batch["data_lag"]
        w_t = train_batch["weights"]
        w_lag = train_batch["weights_lag"]
        # =================forward====================
        f_t = self.forward_nn(x_t)
        f_lag = self.forward_nn(x_lag)
        # ===================tica=====================
        eigvals, _ = self.tica.compute(
            data=[f_t, f_lag], weights=[w_t, w_lag], save_params=True
        )
        # ===================loss=====================
        loss = self.loss_fn(eigvals)
        # ====================log=====================
        name = "train" if self.training else "valid"
        loss_dict = {f"{name}_loss": loss}
        eig_dict = {f"{name}_eigval_{i+1}": eigvals[i] for i in range(len(eigvals))}
        self.log_dict(dict(loss_dict, **eig_dict), on_step=True, on_epoch=True)
        return loss


def test_deep_tica():
    # tests
    import numpy as np
    from mlcolvar.data import DictModule
    from mlcolvar.utils.timelagged import create_timelagged_dataset

    # create dataset
    X = np.loadtxt("mlcolvar/tests/data/mb-mcmc.dat")
    X = torch.Tensor(X)
    dataset = create_timelagged_dataset(X, lag_time=1)
    datamodule = DictModule(dataset, batch_size=10000)

    # create cv
    layers = [2, 10, 10, 2]
    model = DeepTICA(layers, n_cvs=1)

    # change loss options
    model.loss_fn.mode = "sum2"

    # create trainer and fit
    trainer = lightning.Trainer(
        max_epochs=1, log_every_n_steps=2, logger=None, enable_checkpointing=False
    )
    trainer.fit(model, datamodule)

    model.eval()
    with torch.no_grad():
        s = model(X).numpy()
    print(X.shape, "-->", s.shape)


if __name__ == "__main__":
    test_deep_tica()



"""Principal component analysis"""

import torch

from mlcolvar.core.stats import Stats

__all__ = ["PCA"]


class PCA(Stats):
    """
    Principal Component Analysis class.

    Attributes
    ----------
    in_features : int
        Number of features
    out_features : int
        Number of output features
    evals : torch.Tensor
        PCA eigenvalues
    evecs : torch.Tensor
        PCA eigenvectors
    """

    def __init__(self, in_features: int, out_features: int = None):
        """
        Initialize a PCA object. If out_features (<in_features) is given a low rank
        approximation will be performed.

        Parameters
        ----------
        in_features : int
            number of input features
        out_features : int, optional
            number of principal components, by default None (equal to in_features)
        """
        super().__init__()

        # Save attributes
        self.in_features = in_features
        self.out_features = in_features if out_features is None else out_features

        # create eigenvector buffer
        self.register_buffer("evecs", torch.eye(in_features, self.out_features))

        # initialize other attributes
        self.evals = None

    def extra_repr(self) -> str:
        repr = f"in_features={self.in_features}, out_features={self.out_features}"
        return repr

    def compute(self, X, save_params=True, **kwargs):
        """
        Compute PCA eigenvalues and eigenvectors via torch.pca_lowrank method.

        Parameters
        ----------
        X : array-like of shape (n_samples, in_features)
            Training data.
        save_params: bool, optional
            Whether to store parameters in model
        kwargs: optional
            Keyword arguments passed to torch.pca_lowrank (e.g. center)

        Returns
        -------
        eigvals : torch.Tensor
            PCA eigenvalues
        eigvecs : torch.Tensor
            PCA eigenvectors

        Notes
        -----
        The eigenvecs object which is returned is a matrix whose column eigvecs[:,i] is the eigenvector associated to eigvals[i]
        """

        n, d = X.shape
        U, S, V = torch.pca_lowrank(X, q=self.out_features, **kwargs)

        evals = torch.square(S) / (n - 1)
        evecs = V

        if save_params:
            self.evals = evals
            self.evecs = evecs

        return evals, evecs

    @property
    def explained_variance(self):
        """Explained variance ratio of the principal components selected.

        Returns
        -------
        exp_var, torch.Tensor
            explained variance ratio
        """
        return self.evals[: self.out_features] / self.evals.sum()

    @property
    def cumulative_explained_variance(self):
        """Cumulative explained variance.

        Returns
        -------
        cum_exp_var, torch.Tensor
            cumulative variance ratio
        """
        return torch.cumsum(self.explained_variance[: self.out_features], 0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute linear combination with saved eigenvectors.
        If self.out_features is < then the number of eigenvectors only
        the first out_features components will be used.

        Parameters
        ----------
        x: torch.Tensor
            input

        Returns
        -------
        out : torch.Tensor
            output
        """
        return torch.matmul(x, self.evecs[:, : self.out_features])


def test_pca():
    in_features = 10

    torch.manual_seed(42)
    X = torch.rand(100, in_features) * 100

    # find all components (default)
    pca = PCA(in_features)
    _ = pca.compute(X)
    s = pca(X)
    assert s.shape[1] == in_features
    assert len(pca.explained_variance) == in_features
    assert len(pca.cumulative_explained_variance) == in_features

    # select first n_components after calculation
    n_components = 5
    pca.out_features = n_components
    assert len(pca.explained_variance) == n_components
    assert len(pca.cumulative_explained_variance) == n_components

    # select n_components in init
    pca = PCA(in_features, n_components)
    _ = pca.compute(X)
    s = pca(X)
    assert s.shape[1] == n_components
    assert len(pca.explained_variance) == n_components
    assert len(pca.cumulative_explained_variance) == n_components


if __name__ == "__main__":
    test_pca()



import torch
import lightning
from mlcolvar.cvs import BaseCV
from mlcolvar.core import FeedForward, Normalization
from mlcolvar.data import DictModule
from mlcolvar.core.stats import LDA
from mlcolvar.core.loss import ReduceEigenvaluesLoss

__all__ = ["DeepLDA"]


class DeepLDA(BaseCV, lightning.LightningModule):
    """Deep Linear Discriminant Analysis (Deep-LDA) CV.
    Non-linear generalization of LDA in which a feature map is learned by a neural network optimized
    as to maximize the classes separation. The method is described in [1]_.

    **Data**: for training it requires a DictDataset with the keys 'data' and 'labels'.

    **Loss**: maximize LDA eigenvalues (ReduceEigenvaluesLoss)

    References
    ----------
    .. [1] L. Bonati, V. Rizzi, and M. Parrinello, "Data-driven collective variables for enhanced
        sampling", JPCL 11, 2998–3004 (2020).

    See also
    --------
    mlcolvar.core.stats.LDA
        Linear Discriminant Analysis method
    mlcolvar.core.loss.ReduceEigenvalueLoss
        Eigenvalue reduction to a scalar quantity
    """

    BLOCKS = ["norm_in", "nn", "lda"]

    def __init__(self, layers: list, n_states: int, options: dict = None, **kwargs):
        """
        Define a Deep Linear Discriminant Analysis (Deep-LDA) CV composed by a
        neural network module and a LDA object.
        By default a module standardizing the inputs is also used.

        Parameters
        ----------
        layers : list
            Number of neurons per layer
        n_states : int
            Number of states for the training
        options : dict[str, Any], optional
            Options for the building blocks of the model, by default {}.
            Available blocks: ['norm_in','nn','lda'] .
            Set 'block_name' = None or False to turn off that block
        """
        super().__init__(in_features=layers[0], out_features=layers[-1], **kwargs)

        # =======   LOSS  =======
        # Maximize the sum of all the LDA eigenvalues.
        self.loss_fn = ReduceEigenvaluesLoss(mode="sum")

        # ======= OPTIONS =======
        # parse and sanitize
        options = self.parse_options(options)

        # Save n_states
        self.n_states = n_states

        # ======= BLOCKS =======

        # initialize norm_in
        o = "norm_in"
        if (options[o] is not False) and (options[o] is not None):
            self.norm_in = Normalization(self.in_features, **options[o])

        # initialize nn
        o = "nn"
        self.nn = FeedForward(layers, **options[o])

        # initialize lda
        o = "lda"
        self.lda = LDA(layers[-1], n_states, **options[o])

        # regularization
        self.lorentzian_reg = 40  # == 2/sw_reg, see set_regularization
        self.set_regularization(sw_reg=0.05)

    def forward_nn(self, x: torch.Tensor) -> torch.Tensor:
        if self.norm_in is not None:
            x = self.norm_in(x)
        x = self.nn(x)
        return x

    def set_regularization(self, sw_reg=0.05, lorentzian_reg=None):
        r"""
        Set magnitude of regularizations for the training:
        - add identity matrix multiplied by `sw_reg` to within scatter S_w.
        - add lorentzian regularization to NN outputs with magnitude `lorentzian_reg`

        If `lorentzian_reg` is None, set it equal to `2./sw_reg`.

        Parameters
        ----------
        sw_reg : float
            Regularization value for S_w.
        lorentzian_reg: float
            Regularization for lorentzian on NN outputs.

        Notes
        -----
        These regularizations are described in [1]_.

        - S_w
        .. math:: S_w = S_w + \mathtt{sw_reg}\ \mathbf{1}.

        - Lorentzian

        TODO Add equation

        """
        self.lda.sw_reg = sw_reg
        if lorentzian_reg is None:
            self.lorentzian_reg = 2.0 / sw_reg
        else:
            self.lorentzian_reg = lorentzian_reg

    def regularization_lorentzian(self, H):
        """
        Compute lorentzian regularization on NN outputs.

        Parameters
        ----------
        x : float
            input data
        """
        reg_loss = H.pow(2).sum().div(H.size(0))
        reg_loss_lor = -self.lorentzian_reg / (1 + (reg_loss - 1).pow(2))
        return reg_loss_lor

    def training_step(self, train_batch, batch_idx):
        """Compute and return the training loss and record metrics."""
        # =================get data===================
        x = train_batch["data"]
        y = train_batch["labels"]
        # =================forward====================
        h = self.forward_nn(x)
        # ===================lda======================
        eigvals, _ = self.lda.compute(
            h, y, save_params=True if self.training else False
        )
        # ===================loss=====================
        loss = self.loss_fn(eigvals)
        if self.lorentzian_reg > 0:
            lorentzian_reg = self.regularization_lorentzian(h)
            loss += lorentzian_reg
        # ====================log=====================
        name = "train" if self.training else "valid"
        loss_dict = {f"{name}_loss": loss, f"{name}_lorentzian_reg": lorentzian_reg}
        eig_dict = {f"{name}_eigval_{i+1}": eigvals[i] for i in range(len(eigvals))}
        self.log_dict(dict(loss_dict, **eig_dict), on_step=True, on_epoch=True)
        return loss


def test_deeplda(n_states=2):
    from mlcolvar.data import DictDataset

    in_features, out_features = 2, n_states - 1
    layers = [in_features, 50, 50, out_features]

    # create dataset
    n_points = 500
    X, y = [], []
    for i in range(n_states):
        X.append(
            torch.randn(n_points, in_features) * (i + 1)
            + torch.Tensor([10 * i, (i - 1) * 10])
        )
        y.append(torch.ones(n_points) * i)

    X = torch.cat(X, dim=0)
    y = torch.cat(y, dim=0)

    dataset = DictDataset({"data": X, "labels": y})
    datamodule = DictModule(dataset, lengths=[0.8, 0.2], batch_size=n_states * n_points)

    # initialize CV
    opts = {
        "norm_in": {"mode": "mean_std"},
        "nn": {"activation": "relu"},
        "lda": {},
    }
    model = DeepLDA(layers, n_states, options=opts)

    # create trainer and fit
    trainer = lightning.Trainer(
        max_epochs=1, log_every_n_steps=2, logger=None, enable_checkpointing=False
    )
    trainer.fit(model, datamodule)

    # eval
    model.eval()
    with torch.no_grad():
        s = model(X).numpy()


if __name__ == "__main__":
    test_deeplda(n_states=2)
    test_deeplda(n_states=3)



import pytest
import urllib
from mlcolvar.utils.io import load_dataframe
from mlcolvar.utils.io import test_datasetFromFile

example_files = {
    "str": "mlcolvar/tests/data/state_A.dat",
    "list": ["mlcolvar/tests/data/state_A.dat", "mlcolvar/tests/data/state_B.dat"],
    "url": "https://raw.githubusercontent.com/luigibonati/mlcolvar/main/mlcolvar/tests/data/2d_model/COLVAR_stateA",
}


@pytest.mark.parametrize("file_type", ["str", "list", "url"])
def test_loadDataframe(file_type):
    filename = example_files[file_type]
    if file_type == "url":
        # disable test if connection is not available
        try:
            urllib.request.urlopen(filename)
        except urllib.error.URLError:
            pytest.skip("internet not available")

    df = load_dataframe(filename, start=0, stop=10, stride=1)


if __name__ == "__main__":
    # test_loadDataframe()
    test_datasetFromFile()



#!/usr/bin/env python


# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
Test objects and function in mlcolvar.core.nn.feedforward.
"""


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import pytest
import torch

from mlcolvar.core.nn.feedforward import FeedForward


# =============================================================================
# TESTS
# =============================================================================


@pytest.mark.parametrize(
    "activation",
    [
        "relu",
        "elu",
        "softplus",
        "shifted_softplus",
        ["relu", "tanh", None],
        ["relu", "tanh", "relu"],
    ],
)
@pytest.mark.parametrize(
    "dropout",
    [
        None,
        0.5,
        [0.2, 0.3, None],
        [0.2, 0.3, 0.4],
    ],
)
@pytest.mark.parametrize(
    "batchnorm",
    [
        False,
        True,
        [True, True, False],
        [True, False, True],
    ],
)
@pytest.mark.parametrize("last_layer_activation", [False, True])
def test_feedforward(activation, dropout, batchnorm, last_layer_activation):
    """The model is constructed with the correct layers."""
    in_features = 2
    layers = [in_features, 10, 10, 1]
    max_n_activ_layers = len(layers) - 1

    # Create model.
    model = FeedForward(
        layers=layers,
        activation=activation,
        dropout=dropout,
        batchnorm=batchnorm,
        last_layer_activation=last_layer_activation,
    )

    # Check batchnorm.
    if isinstance(batchnorm, list):
        n_batchnorm = sum(batchnorm)
    elif last_layer_activation:
        n_batchnorm = int(batchnorm) * max_n_activ_layers
    else:
        n_batchnorm = int(batchnorm) * (max_n_activ_layers - 1)
    assert sum(isinstance(l, torch.nn.BatchNorm1d) for l in model.nn) == n_batchnorm

    # Check dropout.
    if isinstance(dropout, list):
        n_dropout = sum(d is not None for d in dropout)
    elif last_layer_activation:
        n_dropout = int(dropout is not None) * max_n_activ_layers
    else:
        n_dropout = int(dropout is not None) * (max_n_activ_layers - 1)
    assert sum(isinstance(l, torch.nn.Dropout) for l in model.nn) == n_dropout

    # Check activation.
    if isinstance(activation, list):
        n_activations = sum(a is not None for a in activation)
    elif last_layer_activation:
        n_activations = int(activation is not None) * max_n_activ_layers
    else:
        n_activations = int(activation is not None) * (max_n_activ_layers - 1)
    n_linear = len(layers) - 1
    assert len(model.nn) == n_linear + n_activations + n_dropout + n_batchnorm

    # Test that the forward doesn't explode.
    batch_size = 2
    x = torch.zeros((batch_size, in_features))
    model(x)



"""Linear discriminant analysis"""

import torch
from .utils import cholesky_eigh

from mlcolvar.core.stats import Stats

__all__ = ["LDA"]


class LDA(Stats):
    """
    Fisher's discriminant class.

    Attributes
    ----------
    in_features : int
        Number of input features
    out_features : int
        Number of ouput features (n_states - 1)
    n_states : int
        Number of states
    evals : torch.Tensor
        LDA eigenvalues
    evecs : torch.Tensor
        LDA eigenvectors
    S_b : torch.Tensor
        Between scatter matrix
    S_w : torch.Tensor
        Within scatter matrix
    sw_reg : float
        Regularization to S_w matrix
    """

    def __init__(self, in_features: int, n_states: int, mode: str = "standard"):
        """
        Initialize a LDA object.

        Parameters
        ----------
        in_features : int
            number of input features
        n_states : int
            number of states
        mode : string
            options: {'standard','harmonic'}, standard or harmonic variant
        """
        super().__init__()

        # Save attributes
        self.n_states = n_states
        self.in_features = in_features
        self.out_features = n_states - 1
        if (mode == "standard") or (mode == "harmonic"):
            self.mode = mode
        else:
            raise ValueError(f"LDA mode should be standard or harmonic, not {mode}.")

        # create eigenvector buffer
        self.register_buffer("evecs", torch.eye(in_features, self.out_features))

        # initialize other attributes
        self.evals = None
        self.S_b = None
        self.S_w = None

        # Regularization
        self.sw_reg = 1e-6

    def extra_repr(self) -> str:
        repr = f"in_features={self.in_features}, out_features={self.out_features}"
        if self.mode == "harmonic":
            repr += f" mode={self.mode}"
        return repr

    def compute(self, X, labels, save_params=True):
        """
        Compute LDA eigenvalues and eigenvectors.
        First compute the scatter matrices S_between (S_b) and S_within (S_w) and then solve the generalized eigenvalue problem.

        Parameters
        ----------
        X : array-like of shape (n_samples, in_features)
            Training data.
        labels : array-like of shape (n_samples,)
            states labels.
        save_params: bool, optional
            Whether to store parameters in model

        Returns
        -------
        eigvals : torch.Tensor
            LDA eigenvalues (n_states-1)
        eigvecs : torch.Tensor
            LDA eigenvectors (n_feature,n_states-1)

        Notes
        -----
        The eigenvecs object which is returned is a matrix whose column eigvecs[:,i] is the eigenvector associated to eigvals[i]
        """

        S_b, S_w = self.compute_scatter_matrices(X, labels, save_params)
        evals, evecs = cholesky_eigh(S_b, S_w, self.sw_reg, n_eig=self.n_states - 1)
        if save_params:
            self.evals = evals
            self.evecs = evecs

        return evals, evecs

    def compute_scatter_matrices(self, X, labels, save_params=True):
        """
        Compute between scatter and within scatter matrices.

        Parameters
        ----------
        X : array-like of shape (n_samples, in_features)
            Training data.
        labels : array-like of shape (n_samples,)
            states labels.
        save_params: bool, optional
            Whether to store parameters in model

        Returns
        -------
        S_b,S_w : arrays of shape (in_features,in_features)
            Between and within scatter matrices
        """
        # device
        device = X.device

        # sizes
        N, d = X.shape
        self.in_features = d

        # states
        states = torch.unique(labels)
        n_states = len(states)

        # mean centered observations for entire population
        X_bar = X - torch.mean(X, 0, True)
        # Total scatter matrix (cov matrix over all observations)
        S_t = X_bar.t().matmul(X_bar) / (N - 1)
        # Define within scatter matrix and compute it
        S_w = torch.Tensor().new_zeros((d, d)).to(device)
        if self.mode == "harmonic":
            S_w_inv = torch.Tensor().new_zeros((d, d)).to(device)
        # Loop over states to compute means and covs
        for i in states:
            # check which elements belong to class i
            X_i = X[torch.nonzero(labels == i).view(-1)]
            # compute mean centered obs of class i
            X_i_bar = X_i - torch.mean(X_i, 0, True)
            # count number of elements
            N_i = X_i.shape[0]
            if N_i == 0:
                continue
            # LDA
            S_w += X_i_bar.t().matmul(X_i_bar) / ((N_i - 1) * n_states)

            # HLDA
            if self.mode == "harmonic":
                inv_i = X_i_bar.t().matmul(X_i_bar) / ((N_i - 1) * n_states)
                S_w_inv += inv_i.inverse()

        if self.mode == "harmonic":
            S_w = S_w_inv.inverse()

        # Compute S_b from total scatter matrix
        S_b = S_t - S_w

        if save_params:
            self.S_b = S_b
            self.S_w = S_w

        return S_b, S_w

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute linear combination with saved eigenvectors

        Parameters
        ----------
        x: torch.Tensor
            input

        Returns
        -------
        out : torch.Tensor
            output
        """
        return torch.matmul(x, self.evecs)


def test_lda():
    in_features = 2
    n_states = 2

    torch.manual_seed(42)
    X = torch.rand(100, in_features) * 100
    y = torch.randint(n_states, (100, 1)).squeeze(1)

    # standard
    lda = LDA(in_features, n_states)
    print(lda)
    S_b, S_w = lda.compute_scatter_matrices(X, y)
    print(S_w, S_b)
    evals, evecs = lda.compute(X, y, True)
    print(lda.S_w.shape, lda.S_b.shape)
    print(evals.shape, evecs.shape)
    s = lda(X)
    print(s.shape)
    assert (s.ndim == 2) and (s.shape[1] == n_states - 1)

    # harmonic variant
    hlda = LDA(in_features, n_states, mode="harmonic")
    print(hlda)
    hlda.compute(X, y)
    s = lda(X)


if __name__ == "__main__":
    test_lda()



import torch
from typing import Union
from warnings import warn

__all__ = ["Statistics", "Inverse"]


class Statistics(object):
    """
    Calculate statistics (running mean and std.dev based on Welford's algorithm, as well as min and max).
    If used with an iterable (such as a dataloader) provides the running estimates.
    To get the dictionary with the results use the .to_dict() method.
    """

    def __init__(self, X: torch.Tensor = None):
        self.count = 0

        self.properties = ["mean", "std", "min", "max"]

        # initialize properties and temp var M2
        for prop in self.properties:
            setattr(self, prop, None)
        setattr(self, "M2", None)

        self.__call__(X)

    def __call__(self, x):
        self.update(x)

    def update(self, x):
        if x is None:
            return

        # get batch size
        ndim = x.ndim
        if ndim == 0:
            x = x.reshape(1, 1)
        elif ndim == 1:
            x = x.unsqueeze(1)

        batch_size = x.shape[0]
        nfeatures = x.shape[1]

        new_count = self.count + batch_size

        # Initialize
        if self.mean is None:
            for prop in ["mean", "M2", "std"]:
                setattr(self, prop, torch.zeros(nfeatures))

        # compute sample mean
        sample_mean = torch.mean(x, dim=0)
        sample_m2 = torch.sum((x - sample_mean) ** 2, dim=0)

        # update stats
        delta = sample_mean - self.mean
        self.mean += delta * batch_size / new_count
        corr = batch_size * self.count / new_count
        self.M2 += sample_m2 + delta**2 * corr
        self.count = new_count
        self.std = torch.sqrt(self.M2 / self.count)

        # compute min/max
        sample_min = torch.min(x, dim=0).values
        sample_max = torch.max(x, dim=0).values

        if self.min is None:
            self.min = sample_min
            self.max = sample_max
        else:
            self.min = torch.min(torch.stack((sample_min, self.min)), dim=0).values
            self.max = torch.max(torch.stack((sample_max, self.max)), dim=0).values

    def to_dict(self) -> dict:
        return {prop: getattr(self, prop) for prop in self.properties}

    def __repr__(self):
        repr = "<Statistics>  "
        for prop in self.properties:
            repr += f"{prop}: {getattr(self,prop).numpy()} "
        return repr


class Inverse(torch.nn.Module):
    "Wrapper to return the inverse method of a module as a torch.nn.Module"

    def __init__(self, module: torch.nn.Module):
        """Return the inverse method of a module as a torch.nn.Module

        Parameters
        ----------
        module : torch.nn.Module
            Module to be inverted
        """
        super().__init__()
        if not hasattr(module, "inverse"):
            raise AttributeError("The given module does not have a 'inverse' method!")
        self.module = module

    def inverse(self, *args, **kwargs):
        return self.module.inverse(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return self.inverse(*args, **kwargs)


def batch_reshape(t: torch.Tensor, size: torch.Size) -> torch.Tensor:
    """Return value reshaped according to size.
    In case of batch unsqueeze and expand along the first dimension.
    For single inputs just pass.

    Parameters
    ----------
        mean and range

    """
    if len(size) == 1:
        return t
    if len(size) == 2:
        batch_size = size[0]
        x_size = size[1]
        t = t.unsqueeze(0).expand(batch_size, x_size)
    else:
        raise ValueError(
            f"Input tensor must of shape (n_features) or (n_batch,n_features), not {size} (len={len(size)})."
        )
    return t


# ================================================================================================
# ======================================== TEST FUNCTIONS ========================================
# ================================================================================================


def test_statistics():
    # create fake data
    X = torch.arange(0, 100)
    X = torch.stack([X + 0.0, X + 100.0, X - 1000.0], dim=1)
    y = X.square().sum(1)
    print("X", X.shape)
    print("y", y.shape)

    # compute stats

    stats = Statistics()
    stats(X)
    print(stats)
    stats.to_dict()

    # create dataloader
    from mlcolvar.data import DictLoader

    loader = DictLoader({"data": X, "target": y}, batch_size=20)

    # compute statistics of a single key of loader
    key = "data"
    stats = Statistics()
    for batch in loader:
        stats.update(batch[key])
    print(stats)

    # compute stats of all keys in dataloader

    # init a statistics object for each key
    stats = {}
    for batch in loader:
        for key in loader.keys:
            # initialize
            if key not in stats:
                stats[key] = Statistics(batch[key])
            # or accumulate
            else:
                stats[key].update(batch[key])

    for key in loader.keys:
        print(key, stats[key])


if __name__ == "__main__":
    test_statistics()



#!/usr/bin/env python

# =============================================================================
# MODULE DOCSTRING
# =============================================================================

"""
PyTorch Lightning DataModule object for DictDatasets.
"""

__all__ = ["DictModule"]


# =============================================================================
# GLOBAL IMPORTS
# =============================================================================

import math
from typing import Sequence, Union, Optional
import warnings

import torch
import numpy as np
import lightning
from torch.utils.data import random_split, Subset
from torch._utils import _accumulate

from mlcolvar.data import DictLoader, DictDataset


# =============================================================================
# DICTIONARY DATAMODULE CLASS
# =============================================================================


class DictModule(lightning.LightningDataModule):
    """Lightning DataModule constructed for :class:`~mlcolvar.data.dataset.DictDataset` .

    The DataModule automatically splits the :class:`~mlcolvar.data.dataset.DictDataset` s
    (using either random or sequential splitting) into training, validation, and (optionally)
    test sets.

    The class can also merge multiple :class:`~mlcolvar.data.dataset.DictDataset`s
    that have different keys (see example below). The datasets must all have the
    same number of samples.

    Examples
    --------

    >>> x = torch.randn((50, 2))
    >>> dataset = DictDataset({'data': x, 'labels': x.square().sum(dim=1)})
    >>> datamodule = DictModule(dataset, lengths=[0.75,0.2,0.05], batch_size=25)

    >>> # This is usually called by PyTorch Lightning.
    >>> datamodule.setup()
    >>> train_loader = datamodule.train_dataloader()
    >>> for batch in train_loader:
    ...     batch_x = batch['data']
    ...     batch_y = batch['labels']
    ...     print(batch_x.shape, batch_y.shape)
    torch.Size([25, 2]) torch.Size([25])
    torch.Size([13, 2]) torch.Size([13])

    >>> val_loader = datamodule.val_dataloader()
    >>> test_loader = datamodule.test_dataloader()

    You can also iterate over multiple datasets. These can have different keys,
    but they must be of the same length.

    >>> dataset2 = DictDataset({'data2': torch.randn(50, 1), 'weights': torch.arange(50)})
    >>> datamodule = DictModule([dataset, dataset2], lengths=[0.8, 0.2], batch_size=5)
    >>> datamodule.setup()
    >>> train_loader = datamodule.train_dataloader()

    >>> # Print first batch.
    >>> for batch in train_loader:
    ...     print('batch dataset0:', list(batch['dataset0'].keys()))
    ...     print('batch dataset1:', list(batch['dataset1'].keys()))
    ...     break
    batch dataset0: ['data', 'labels']
    batch dataset1: ['data2', 'weights']

    """

    def __init__(
        self,
        dataset: DictDataset,
        lengths: Sequence = (0.8, 0.2),
        batch_size: Union[int, Sequence] = 0,
        random_split: bool = True,
        shuffle: Union[bool, Sequence] = True,
        generator: Optional[torch.Generator] = None,
    ):
        """Create a ``DataModule`` wrapping a :class:`~mlcolvar.data.dataset.DictDataset`.

        For the ``batch_size`` and ``shuffle`` parameters, either a single value
        or a list-type of values (with same size as ``lengths``) can be provided.

        Parameters
        ----------
        dataset : DictDataset or Sequence[DictDataset]
            The dataset or a list of datasets. If a list, the datasets can have
            different keys but they must all have the same number of samples.
        lengths : list-like, optional
            Lengths of the training, validation, and (optionally) test datasets.
            This must be a list of (float) fractions summing to 1. The default is
            ``[0.8,0.2]``.
        batch_size : int or list-like, optional
            Batch size, by default 0 (== ``len(dataset)``).
        random_split: bool, optional
            Whether to randomly split train/valid/test or sequentially, by default ``True``.
        shuffle : int or list-like, optional
            Whether to shuffle the batches in the ``DataLoader``, by default ``True``.
        generator : torch.Generator, optional
            Set random generator for reproducibility, by default ``None``.

        See Also
        --------
        :class:`~mlcolvar.data.dataloader.DictLoader`
            The PyTorch loader built by the data module.

        """
        super().__init__()
        self.dataset = dataset
        self.lengths = lengths
        self.generator = generator

        # Keeping this private for now. Changing it at runtime would
        # require changing dataset_split and the dataloaders.
        self._random_split = random_split

        # Make sure batch_size and shuffle are lists.
        if isinstance(batch_size, int):
            self.batch_size = [batch_size for _ in lengths]
        else:
            self.batch_size = batch_size
        if isinstance(shuffle, bool):
            self.shuffle = [shuffle for _ in lengths]
        else:
            self.shuffle = shuffle

        # This is initialized in setup().
        self._dataset_split = None

        # dataloaders
        self.train_loader = None
        self.valid_loader = None
        self.test_loader = None

    def setup(self, stage: Optional[str] = None):
        if self._dataset_split is None:
            if isinstance(self.dataset, DictDataset):
                self._dataset_split = self._split(self.dataset)
            else:  # List of datasets.
                dataset_split = [self._split(d) for d in self.dataset]
                # Shape (n_datasets, n_loaders) -> (n_loaders, n_datasets)
                self._dataset_split = list(map(list, zip(*dataset_split)))

    def train_dataloader(self):
        """Return training dataloader."""
        self._check_setup()
        if self.train_loader is None:
            self.train_loader = DictLoader(
                self._dataset_split[0],
                batch_size=self.batch_size[0],
                shuffle=self.shuffle[0],
            )
        return self.train_loader

    def val_dataloader(self):
        """Return validation dataloader."""
        self._check_setup()
        if len(self.lengths) < 2:
            raise NotImplementedError(
                "Validation dataset not available, you need to pass two lengths to datamodule."
            )
        if self.valid_loader is None:
            self.valid_loader = DictLoader(
                self._dataset_split[1],
                batch_size=self.batch_size[1],
                shuffle=self.shuffle[1],
            )
        return self.valid_loader

    def test_dataloader(self):
        """Return test dataloader."""
        self._check_setup()
        if len(self.lengths) < 3:
            raise NotImplementedError(
                "Test dataset not available, you need to pass three lengths to datamodule."
            )
        if self.test_loader is None:
            self.test_loader = DictLoader(
                self._dataset_split[2],
                batch_size=self.batch_size[2],
                shuffle=self.shuffle[2],
            )
        return self.test_loader

    def predict_dataloader(self):
        raise NotImplementedError()

    def teardown(self, stage: str):
        pass

    def __repr__(self) -> str:
        string = f"DictModule(dataset -> {self.dataset.__repr__()}"
        string += f",\n\t\t     train_loader -> DictLoader(length={self.lengths[0]}, batch_size={self.batch_size[0]}, shuffle={self.shuffle[0]})"
        if len(self.lengths) >= 2:
            string += f",\n\t\t     valid_loader -> DictLoader(length={self.lengths[1]}, batch_size={self.batch_size[1]}, shuffle={self.shuffle[1]})"
        if len(self.lengths) >= 3:
            string += f",\n\t\t\ttest_loader =DictLoader(length={self.lengths[2]}, batch_size={self.batch_size[2]}, shuffle={self.shuffle[2]})"
        string += f")"
        return string

    def _split(self, dataset):
        """Perform the random or sequential spliting of a single dataset.

        Returns a list of Subset[DictDataset] objects.
        """
        if self._random_split:
            dataset_split = random_split(
                dataset, self.lengths, generator=self.generator
            )
        else:
            dataset_split = sequential_split(dataset, self.lengths)
        return dataset_split

    def _check_setup(self):
        """Raise an error if setup() has not been called."""
        if self._dataset_split is None:
            raise AttributeError(
                "The datamodule has not been set up yet. To get the dataloaders "
                "outside a Lightning trainer please call .setup() first."
            )


def sequential_split(dataset, lengths: Sequence) -> list:
    """
    Sequentially split a dataset into non-overlapping new datasets of given lengths.

    The behavior is the same as torch.utils.data.dataset.random_split.

    If a list of fractions that sum up to 1 is given,
    the lengths will be computed automatically as
    floor(frac * len(dataset)) for each fraction provided.

    After computing the lengths, if there are any remainders, 1 count will be
    distributed in round-robin fashion to the lengths
    until there are no remainders left.
    """

    if math.isclose(sum(lengths), 1) and sum(lengths) <= 1:
        subset_lengths = []
        for i, frac in enumerate(lengths):
            if frac < 0 or frac > 1:
                raise ValueError(f"Fraction at index {i} is not between 0 and 1")
            n_items_in_split = int(
                math.floor(len(dataset) * frac)  # type: ignore[arg-type]
            )
            subset_lengths.append(n_items_in_split)
        remainder = len(dataset) - sum(subset_lengths)  # type: ignore[arg-type]
        # add 1 to all the lengths in round-robin fashion until the remainder is 0
        for i in range(remainder):
            idx_to_add_at = i % len(subset_lengths)
            subset_lengths[idx_to_add_at] += 1
        lengths = subset_lengths
        for i, length in enumerate(lengths):
            if length == 0:
                warnings.warn(
                    f"Length of split at index {i} is 0. "
                    f"This might result in an empty dataset."
                )

        # Cannot verify that dataset is Sized
        if sum(lengths) != len(dataset):  # type: ignore[arg-type]
            raise ValueError(
                "Sum of input lengths does not equal the length of the input dataset!"
            )

        # LB change: do sequential rather then random splitting
        return [
            Subset(dataset, np.arange(offset - length, offset))
            for offset, length in zip(_accumulate(lengths), lengths)
        ]
    else:
        raise NotImplementedError("The lengths must sum to 1.")


if __name__ == "__main__":
    import doctest

    doctest.testmod()
